@@ -31,48 +31,44 @@
 
 #include "config.h"
 
+#include <string.h>
 #include <stdio.h>
 #include <errno.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <signal.h>
-#include <sys/types.h>
 #include <sys/wait.h>
 
-#include "lnav_log.hh"
+#include "base/lnav_log.hh"
+#include "base/string_util.hh"
 #include "lnav_util.hh"
 #include "grep_proc.hh"
-
-#include "time_T.hh"
+#include "vis_line.hh"
 
 using namespace std;
 
-grep_proc::grep_proc(pcre *code, grep_proc_source &gps)
+template<typename LineType>
+grep_proc<LineType>::grep_proc(pcre *code, grep_proc_source<LineType> &gps)
     : gp_pcre(code),
-      gp_code(code),
-      gp_source(gps),
-      gp_pipe_offset(0),
-      gp_child(-1),
-      gp_child_started(false),
-      gp_last_line(0),
-      gp_sink(NULL),
-      gp_control(NULL)
+      gp_source(gps)
 {
     require(this->invariant());
+
+    gps.register_proc(this);
 }
 
-grep_proc::~grep_proc()
+template<typename LineType>
+grep_proc<LineType>::~grep_proc()
 {
-    this->gp_queue.clear();
-    this->cleanup();
+    this->invalidate();
 }
 
-void grep_proc::handle_match(int line,
-                             string &line_value,
-                             int off,
-                             int *matches,
-                             int count)
+template<typename LineType>
+void grep_proc<LineType>::handle_match(int line,
+                                       string &line_value,
+                                       int off,
+                                       int *matches,
+                                       int count)
 {
     int lpc;
 
@@ -94,7 +90,8 @@ void grep_proc::handle_match(int line,
     }
 }
 
-void grep_proc::start(void)
+template<typename LineType>
+void grep_proc<LineType>::start()
 {
     require(this->invariant());
 
@@ -131,8 +128,9 @@ void grep_proc::start(void)
         log_perror(fcntl(err_pipe.read_end(), F_SETFL, O_NONBLOCK));
         log_perror(fcntl(err_pipe.read_end(), F_SETFD, 1));
         require(this->gp_err_pipe.get() == -1);
-        this->gp_err_pipe      = err_pipe.read_end();
+        this->gp_err_pipe = std::move(err_pipe.read_end());
         this->gp_child_started = true;
+        this->gp_child_queue_size = this->gp_queue.size();
 
         this->gp_queue.clear();
         return;
@@ -154,31 +152,28 @@ void grep_proc::start(void)
     _exit(0);
 }
 
-void grep_proc::child_loop(void)
+template<typename LineType>
+void grep_proc<LineType>::child_loop()
 {
     char   outbuf[BUFSIZ * 2];
     string line_value;
 
-    *stdout = *(fdopen(STDOUT_FILENO, "w"));
     /* Make sure buffering is on, not sure of the state in the parent. */
     if (setvbuf(stdout, outbuf, _IOFBF, BUFSIZ * 2) < 0) {
         perror("setvbuf");
     }
+    lnav_log_file = fopen("/tmp/lnav.grep.err", "a");
     line_value.reserve(BUFSIZ * 2);
     while (!this->gp_queue.empty()) {
-        grep_line_t start_line = this->gp_queue.front().first;
-        grep_line_t stop_line  = this->gp_queue.front().second;
+        LineType start_line = this->gp_queue.front().first;
+        LineType stop_line  = this->gp_queue.front().second;
         bool        done       = false;
-        int         line;
+        LineType line;
 
         this->gp_queue.pop_front();
-        if (start_line == -1) {
-            start_line = this->gp_highest_line;
-            log_debug("highest %d", start_line);
-        }
-        for (line = start_line;
-             (stop_line == -1 || line < stop_line) && !done;
-             line++) {
+        for (line = this->gp_source.grep_initial_line(start_line, this->gp_highest_line);
+             line != -1 && (stop_line == -1 || line < stop_line) && !done;
+             this->gp_source.grep_next_line(line)) {
             line_value.clear();
             done = !this->gp_source.grep_value_for_line(line, line_value);
             if (!done) {
@@ -190,7 +185,7 @@ void grep_proc::child_loop(void)
                     pcre_context::capture_t *m;
 
                     if (pi.pi_offset == 0) {
-                        fprintf(stdout, "%d\n", line);
+                        fprintf(stdout, "%d\n", (int) line);
                     }
                     m = pc.all();
                     fprintf(stdout, "[%d:%d]\n", m->c_begin, m->c_end);
@@ -235,7 +230,8 @@ void grep_proc::child_loop(void)
     }
 }
 
-void grep_proc::cleanup(void)
+template<typename LineType>
+void grep_proc<LineType>::cleanup()
 {
     if (this->gp_child != -1 && this->gp_child != 0) {
         int status = 0;
@@ -249,7 +245,9 @@ void grep_proc::cleanup(void)
         this->gp_child_started = false;
 
         if (this->gp_sink) {
-            this->gp_sink->grep_end(*this);
+            for (size_t lpc = 0; lpc < this->gp_child_queue_size; lpc++) {
+                this->gp_sink->grep_end(*this);
+            }
         }
     }
 
@@ -257,7 +255,7 @@ void grep_proc::cleanup(void)
         this->gp_err_pipe.reset();
     }
 
-    this->gp_pipe_offset = 0;
+    this->gp_pipe_range.clear();
     this->gp_line_buffer.reset();
 
     ensure(this->invariant());
@@ -267,14 +265,14 @@ void grep_proc::cleanup(void)
     }
 }
 
-void grep_proc::dispatch_line(char *line)
+template<typename LineType>
+void grep_proc<LineType>::dispatch_line(char *line)
 {
     int start, end, capture_start;
 
-    require(line != NULL);
+    require(line != nullptr);
 
     if (sscanf(line, "h%d", this->gp_highest_line.out()) == 1) {
-
     } else if (sscanf(line, "%d", this->gp_last_line.out()) == 1) {
         /* Starting a new line with matches. */
         ensure(this->gp_last_line >= 0);
@@ -284,7 +282,7 @@ void grep_proc::dispatch_line(char *line)
         require(end >= 0);
 
         /* Pass the match offsets to the sink delegate. */
-        if (this->gp_sink != NULL) {
+        if (this->gp_sink != nullptr) {
             this->gp_sink->grep_match(*this, this->gp_last_line, start, end);
         }
     }
@@ -293,17 +291,17 @@ void grep_proc::dispatch_line(char *line)
         require(end >= 0);
 
         /* Pass the captured strings to the sink delegate. */
-        if (this->gp_sink != NULL) {
+        if (this->gp_sink != nullptr) {
             this->gp_sink->grep_capture(*this,
                                         this->gp_last_line,
                                         start,
                                         end,
                                         start < 0 ?
-                                        NULL : &line[capture_start]);
+                                        nullptr : &line[capture_start]);
         }
     }
     else if (line[0] == '/') {
-        if (this->gp_sink != NULL) {
+        if (this->gp_sink != nullptr) {
             this->gp_sink->grep_match_end(*this, this->gp_last_line);
         }
     }
@@ -312,7 +310,8 @@ void grep_proc::dispatch_line(char *line)
     }
 }
 
-void grep_proc::check_poll_set(const std::vector<struct pollfd> &pollfds)
+template<typename LineType>
+void grep_proc<LineType>::check_poll_set(const std::vector<struct pollfd> &pollfds)
 {
     require(this->invariant());
 
@@ -328,10 +327,10 @@ void grep_proc::check_poll_set(const std::vector<struct pollfd> &pollfds)
             if (strncmp(buffer, PREFIX, strlen(PREFIX)) == 0) {
                 char *lf;
 
-                if ((lf = strchr(buffer, '\n')) != NULL) {
+                if ((lf = strchr(buffer, '\n')) != nullptr) {
                     *lf = '\0';
                 }
-                if (this->gp_control != NULL) {
+                if (this->gp_control != nullptr) {
                     this->gp_control->grep_error(&buffer[strlen(PREFIX)]);
                 }
             }
@@ -346,22 +345,43 @@ void grep_proc::check_poll_set(const std::vector<struct pollfd> &pollfds)
         try {
             static const int MAX_LOOPS = 100;
 
-            int    loop_count = 0;
-            line_value lv;
+            int loop_count = 0;
+            bool drained = false;
+
+            while (loop_count < MAX_LOOPS) {
+                auto load_result = this->gp_line_buffer
+                    .load_next_line(this->gp_pipe_range);
+
+                if (load_result.isErr()) {
+                    break;
+                }
+
+                auto li = load_result.unwrap();
+
+                if (li.li_file_range.empty()) {
+                    drained = true;
+                    break;
+                }
+
+                this->gp_pipe_range = li.li_file_range;
+                this->gp_line_buffer.read_range(li.li_file_range).then([this](auto sbr) {
+                    auto_mem<char> buf;
+
+                    buf = (char *) malloc(sbr.length() + 1);
+                    sbr.rtrim(is_line_ending);
+                    memcpy(buf, sbr.get_data(), sbr.length());
+                    buf[sbr.length()] = '\0';
+                    this->dispatch_line(buf);
+                });
 
-            while ((loop_count < MAX_LOOPS) &&
-                   (this->gp_line_buffer.read_line(this->gp_pipe_offset, lv))) {
-                lv.terminate();
-                this->dispatch_line(lv.lv_start);
                 loop_count += 1;
             }
 
-            if (this->gp_sink != NULL) {
+            if (this->gp_sink != nullptr) {
                 this->gp_sink->grep_end_batch(*this);
             }
 
-            if ((off_t) this->gp_line_buffer.get_file_size() ==
-                this->gp_pipe_offset) {
+            if (drained && this->gp_line_buffer.is_pipe_closed()) {
                 this->cleanup();
             }
         }
@@ -372,3 +392,18 @@ void grep_proc::check_poll_set(const std::vector<struct pollfd> &pollfds)
 
     ensure(this->invariant());
 }
+
+template<typename LineType>
+grep_proc<LineType> &grep_proc<LineType>::invalidate()
+{
+    if (this->gp_sink) {
+        for (size_t lpc = 0; lpc < this->gp_queue.size(); lpc++) {
+            this->gp_sink->grep_end(*this);
+        }
+    }
+    this->gp_queue.clear();
+    this->cleanup();
+    return *this;
+}
+
+template class grep_proc<vis_line_t>;
