@@ -32,10 +32,11 @@
 #include "config.h"
 
 #include <time.h>
-#include <math.h>
 #include <sys/time.h>
 
-#include "lnav_log.hh"
+#include <cmath>
+
+#include "base/lnav_log.hh"
 #include "listview_curses.hh"
 
 using namespace std;
@@ -43,54 +44,38 @@ using namespace std;
 list_gutter_source listview_curses::DEFAULT_GUTTER_SOURCE;
 
 listview_curses::listview_curses()
-    : lv_source(NULL),
-      lv_overlay_source(NULL),
-      lv_window(NULL),
-      lv_x(0),
-      lv_y(0),
-      lv_top(0),
-      lv_left(0),
-      lv_height(0),
-      lv_needs_update(true),
-      lv_overlay_needs_update(true),
-      lv_show_scrollbar(true),
-      lv_show_bottom_border(false),
-      lv_gutter_source(&DEFAULT_GUTTER_SOURCE),
-      lv_word_wrap(false),
-      lv_scroll_accel(0),
-      lv_scroll_velo(0),
-      lv_mouse_y(-1),
-      lv_mouse_mode(LV_MODE_NONE),
-      lv_tail_space(1)
-{ }
-
-listview_curses::~listview_curses()
-{ }
-
-void listview_curses::reload_data(void)
+    : lv_scroll(noop_func{})
 {
-    if (this->lv_source == NULL) {
-        this->lv_top  = vis_line_t(0);
+}
+
+void listview_curses::reload_data()
+{
+    if (this->lv_source == nullptr) {
+        this->lv_top  = 0_vl;
         this->lv_left = 0;
     }
-    else if (this->lv_top >= this->get_inner_height()) {
-        this->lv_top = max(vis_line_t(0),
-                           vis_line_t(this->get_inner_height() - 1));
+    else {
+        if (this->lv_top >= this->get_inner_height()) {
+            this->lv_top = max(0_vl, vis_line_t(this->get_inner_height() - 1));
+        }
+        if (this->get_inner_height() == 0) {
+            this->lv_selection = 0_vl;
+        } else if (this->lv_selection >= this->get_inner_height()) {
+            this->lv_selection = this->get_inner_height() - 1_vl;
+        }
     }
-    this->lv_needs_update = true;
+    this->vc_needs_update = true;
 }
 
 bool listview_curses::handle_key(int ch)
 {
-    for (list<list_input_delegate *>::iterator iter = this->lv_input_delegates.begin();
-         iter != this->lv_input_delegates.end();
-         ++iter) {
-        if ((*iter)->list_input_handle_key(*this, ch)) {
+    for (auto &lv_input_delegate : this->lv_input_delegates) {
+        if (lv_input_delegate->list_input_handle_key(*this, ch)) {
             return true;
         }
     }
 
-    vis_line_t height(0);
+    auto height = 0_vl;
 
     unsigned long width;
     bool          retval = true;
@@ -118,28 +103,36 @@ bool listview_curses::handle_key(int ch)
     case '\r':
     case 'j':
     case KEY_DOWN:
-        this->shift_top(vis_line_t(1));
+        if (this->is_selectable()) {
+            this->shift_selection(1);
+        } else {
+            this->shift_top(1_vl);
+        }
         break;
 
     case 'k':
     case KEY_UP:
-        this->shift_top(vis_line_t(-1));
+        if (this->is_selectable()) {
+            this->shift_selection(-1);
+        } else {
+            this->shift_top(-1_vl);
+        }
         break;
 
     case 'b':
     case KEY_BACKSPACE:
     case KEY_PPAGE:
-        this->shift_top(-(this->rows_available(this->lv_top, RD_UP) - vis_line_t(1)));
+        this->shift_top(-(this->rows_available(this->lv_top, RD_UP) - 1_vl));
         break;
 
     case ' ':
     case KEY_NPAGE:
-        this->shift_top(this->rows_available(this->lv_top, RD_DOWN) - vis_line_t(1));
+        this->shift_top(this->rows_available(this->lv_top, RD_DOWN) - 1_vl);
         break;
 
     case 'g':
     case KEY_HOME:
-        this->set_top(vis_line_t(0));
+        this->set_top(0_vl);
         break;
 
     case 'G':
@@ -147,12 +140,17 @@ bool listview_curses::handle_key(int ch)
             vis_line_t last_line(this->get_inner_height() - 1);
             vis_line_t tail_bottom(this->get_top_for_last_row());
 
-            if (this->get_top() == last_line)
+            if (this->is_selectable()) {
+                this->set_selection(last_line);
+            } else if (this->get_top() == last_line) {
                 this->set_top(tail_bottom);
-            else if (tail_bottom <= this->get_top())
+            }
+            else if (tail_bottom <= this->get_top()) {
                 this->set_top(last_line);
-            else
+            }
+            else {
                 this->set_top(tail_bottom);
+            }
         }
         break;
 
@@ -181,27 +179,29 @@ bool listview_curses::handle_key(int ch)
     return retval;
 }
 
-void listview_curses::do_update(void)
+void listview_curses::do_update()
 {
-    if (this->lv_window == NULL || this->lv_height == 0) {
+    if (this->lv_window == nullptr || this->lv_height == 0) {
+        view_curses::do_update();
         return;
     }
 
-    if (this->lv_needs_update) {
-        vis_line_t        y(this->lv_y), height, bottom, row;
+    if (this->vc_needs_update) {
+        view_colors &vc = view_colors::singleton();
+        vis_line_t        height, row;
         attr_line_t       overlay_line;
-        vis_line_t        overlay_height(0);
         struct line_range lr;
         unsigned long     width, wrap_width;
         size_t            row_count;
-
-        if (this->lv_overlay_source != NULL) {
-            overlay_height = vis_line_t(
-                this->lv_overlay_source->list_overlay_count(*this));
-        }
+        int y = this->lv_y, bottom;
+        attr_t role_attrs = vc.attrs_for_role(this->vc_default_role);
 
         this->get_dimensions(height, width);
 
+        if (this->vc_width > 0) {
+            width = std::min((unsigned long) this->vc_width, width);
+        }
+
         wrap_width = width - (this->lv_word_wrap ? 1 : this->lv_show_scrollbar ? 1 : 0);
 
         row_count = this->get_inner_height();
@@ -212,12 +212,13 @@ void listview_curses::do_update(void)
         while (y < bottom) {
             lr.lr_start = this->lv_left;
             lr.lr_end   = this->lv_left + wrap_width;
-            if (this->lv_overlay_source != NULL &&
+            if (this->lv_overlay_source != nullptr &&
                 this->lv_overlay_source->list_value_for_overlay(
                     *this,
-                    y - vis_line_t(this->lv_y),
+                    y - this->lv_y, bottom - this->lv_y,
+                    row,
                     overlay_line)) {
-                this->mvwattrline(this->lv_window, y, this->lv_x, overlay_line, lr);
+                mvwattrline(this->lv_window, y, this->lv_x, overlay_line, lr);
                 overlay_line.clear();
                 ++y;
             }
@@ -225,10 +226,10 @@ void listview_curses::do_update(void)
                 attr_line_t &al = rows[row - this->lv_top];
 
                 do {
-                    this->mvwattrline(this->lv_window, y, this->lv_x, al, lr);
+                    mvwattrline(this->lv_window, y, this->lv_x, al, lr,
+                                this->vc_default_role);
                     if (this->lv_word_wrap) {
-                        wmove(this->lv_window, y, wrap_width);
-                        wclrtoeol(this->lv_window);
+                        mvwhline(this->lv_window, y, this->lv_x + wrap_width, ' ', width - wrap_width);
                     }
                     lr.lr_start += wrap_width;
                     lr.lr_end += wrap_width;
@@ -237,14 +238,14 @@ void listview_curses::do_update(void)
                 ++row;
             }
             else {
-                wmove(this->lv_window, y, this->lv_x);
-                wclrtoeol(this->lv_window);
+                wattron(this->lv_window, role_attrs);
+                mvwhline(this->lv_window, y, this->lv_x, ' ', width);
+                wattroff(this->lv_window, role_attrs);
                 ++y;
             }
         }
 
         if (this->lv_show_scrollbar) {
-            view_colors &vc = view_colors::singleton();
             double progress = 1.0;
             double coverage = 1.0;
             double adjusted_height = (double)row_count / (double)height;
@@ -255,17 +256,15 @@ void listview_curses::do_update(void)
                 coverage = (double)height / (double)row_count;
             }
 
-            y = vis_line_t(this->lv_y) +
-                vis_line_t((int)(progress * (double)height));
-            lines = y + min(height, vis_line_t(
-                                (int)(coverage * (double)height)));
+            y = this->lv_y + (int)(progress * (double)height);
+            lines = vis_line_t(y + min((int) height, (int)(coverage * (double)height)));
 
             for (unsigned int gutter_y = this->lv_y;
                  gutter_y < (this->lv_y + height);
                  gutter_y++) {
                 int range_start = 0, range_end;
-                view_colors::role_t role = view_colors::VCR_TEXT;
-                view_colors::role_t bar_role = view_colors::VCR_STATUS;
+                view_colors::role_t role = this->vc_default_role;
+                view_colors::role_t bar_role = view_colors::VCR_SCROLLBAR;
                 int attrs;
                 chtype ch = ACS_VLINE;
 
@@ -281,19 +280,29 @@ void listview_curses::do_update(void)
                 }
                 attrs = vc.attrs_for_role(role);
                 wattron(this->lv_window, attrs);
-                mvwaddch(this->lv_window, gutter_y, width - 1, ch);
+                mvwaddch(this->lv_window, gutter_y, this->lv_x + width - 1, ch);
                 wattroff(this->lv_window, attrs);
             }
             wmove(this->lv_window, this->lv_y + height - 1, this->lv_x);
         }
 
         if (this->lv_show_bottom_border) {
-            mvwchgat(this->lv_window,
-                this->lv_y + height - 1, this->lv_x, width - 1, A_UNDERLINE, 0, NULL);
+            cchar_t row_ch[width];
+            int y = this->lv_y + height - 1;
+
+            mvwin_wchnstr(this->lv_window, y, this->lv_x, row_ch, width - 1);
+            for (unsigned long lpc = 0; lpc < width - 1; lpc++) {
+                row_ch[lpc].attr |= A_UNDERLINE;
+            }
+            mvwadd_wchnstr(this->lv_window, y, this->lv_x, row_ch, width - 1);
         }
 
-        this->lv_needs_update = false;
+        this->vc_needs_update = false;
     }
+
+    view_curses::do_update();
+
+#if 0
     else if (this->lv_overlay_needs_update && this->lv_overlay_source != NULL) {
         vis_line_t y(this->lv_y), height, bottom;
         attr_line_t overlay_line;
@@ -319,11 +328,26 @@ void listview_curses::do_update(void)
             ++y;
         }
     }
+#endif
+}
+
+void listview_curses::shift_selection(int offset)
+{
+    vis_line_t new_selection = this->lv_selection + vis_line_t(offset);
+
+    if (new_selection >= 0_vl &&
+        new_selection < this->get_inner_height()) {
+        this->set_selection(new_selection);
+        this->scroll_selection_into_view();
+    } else if (!alerter::singleton().chime()) {
+        // XXX Disabling for now...
+        // this->delegate_scroll_out();
+    }
 }
 
 static int scroll_polarity(mouse_button_t button)
 {
-    return button == BUTTON_SCROLL_UP ? -1 : 1;
+    return button == mouse_button_t::BUTTON_SCROLL_UP ? -1 : 1;
 }
 
 bool listview_curses::handle_mouse(mouse_event &me)
@@ -337,33 +361,33 @@ bool listview_curses::handle_mouse(mouse_event &me)
     inner_height = this->get_inner_height();
 
     switch (me.me_button) {
-    case BUTTON_SCROLL_UP:
-    case BUTTON_SCROLL_DOWN:
-        if (diff.tv_sec > 0 || diff.tv_usec > 80000) {
-            this->lv_scroll_accel = 1;
-            this->lv_scroll_velo = 0;
-        }
-        else {
-            this->lv_scroll_accel += 2;
-        }
-        this->lv_scroll_velo += this->lv_scroll_accel;
+        case mouse_button_t::BUTTON_SCROLL_UP:
+        case mouse_button_t::BUTTON_SCROLL_DOWN:
+            if (diff.tv_sec > 0 || diff.tv_usec > 80000) {
+                this->lv_scroll_accel = 1;
+                this->lv_scroll_velo = 0;
+            }
+            else {
+                this->lv_scroll_accel += 2;
+            }
+            this->lv_scroll_velo += this->lv_scroll_accel;
 
-        this->shift_top(vis_line_t(scroll_polarity(me.me_button) *
-                                   this->lv_scroll_velo),
-                        true);
-        break;
-    default:
-        break;
+            this->shift_top(vis_line_t(scroll_polarity(me.me_button) *
+                                       this->lv_scroll_velo),
+                            true);
+            break;
+        default:
+            break;
     }
     this->lv_mouse_time = me.me_time;
 
-    if (me.me_button != BUTTON_LEFT ||
+    if (me.me_button != mouse_button_t::BUTTON_LEFT ||
         inner_height == 0 ||
         (this->lv_mouse_mode != LV_MODE_DRAG && me.me_x < (int)(width - 2))) {
         return false;
     }
 
-    if (me.me_state == BUTTON_STATE_RELEASED) {
+    if (me.me_state == mouse_button_state_t::BUTTON_STATE_RELEASED) {
         this->lv_mouse_y = -1;
         this->lv_mouse_mode = LV_MODE_NONE;
         return true;
@@ -421,3 +445,35 @@ bool listview_curses::handle_mouse(mouse_event &me)
 
     return true;
 }
+
+void listview_curses::set_top(vis_line_t top, bool suppress_flash)
+{
+    auto inner_height = this->get_inner_height();
+
+    if (inner_height > 0 && top >= inner_height) {
+        top = vis_line_t(inner_height - 1);
+    }
+    if (top < 0 || (top > 0 && top >= inner_height)) {
+        if (suppress_flash == false) {
+            alerter::singleton().chime();
+        }
+    }
+    else if (this->lv_top != top) {
+        this->lv_top = top;
+        if (this->lv_selectable) {
+            if (this->lv_selection < top) {
+                this->lv_selection = top;
+            } else {
+                auto bot = this->get_bottom();
+
+                if (bot != -1_vl) {
+                    if (this->lv_selection > bot) {
+                        this->lv_selection = bot;
+                    }
+                }
+            }
+        }
+        this->invoke_scroll();
+        this->set_needs_update();
+    }
+}
