@@ -29,9 +29,12 @@
 
 #include "config.h"
 
-#include "lnav_log.hh"
+#include "base/lnav_log.hh"
+#include "base/string_util.hh"
 #include "sql_util.hh"
 #include "log_vtab_impl.hh"
+#include "yajlpp/yajlpp_def.hh"
+#include "vtab_module.hh"
 
 #include "logfile_sub_source.hh"
 
@@ -39,7 +42,31 @@ using namespace std;
 
 static struct log_cursor log_cursor_latest;
 
-sql_progress_callback_t log_vtab_progress_callback;
+struct _log_vtab_data log_vtab_data;
+
+static const char *LOG_COLUMNS = R"(  (
+  log_line        INTEGER  PRIMARY KEY,            -- The line number for the log message
+  log_part        TEXT     COLLATE naturalnocase,  -- The partition the message is in
+  log_time        DATETIME,                        -- The adjusted timestamp for the log message
+  log_actual_time DATETIME HIDDEN,                 -- The timestamp from the original log file for this message
+  log_idle_msecs  INTEGER,                         -- The difference in time between this messages and the previous
+  log_level       TEXT     COLLATE loglevel,       -- The log message level
+  log_mark        BOOLEAN,                         -- True if the log message was marked
+  log_comment     TEXT,                            -- The comment for this message
+  log_tags        TEXT,                            -- A JSON list of tags for this message
+  log_filters     TEXT,                            -- A JSON list of filter IDs that matched this message
+  -- BEGIN Format-specific fields:
+)";
+
+static const char *LOG_FOOTER_COLUMNS = R"(
+  -- END Format-specific fields
+  log_time_msecs  INTEGER HIDDEN,                    -- The adjusted timestamp for the log message as the number of milliseconds from the epoch
+  log_path        TEXT HIDDEN COLLATE naturalnocase, -- The path to the log file this message is from
+  log_text        TEXT HIDDEN,                       -- The full text of the log message
+  log_body        TEXT HIDDEN,                       -- The body of the log message
+  log_raw_text    TEXT HIDDEN                        -- The raw text from the log file
+);
+)";
 
 static const char *type_to_string(int type)
 {
@@ -56,25 +83,17 @@ static const char *type_to_string(int type)
 
     ensure("Invalid sqlite type");
 
-    return NULL;
+    return nullptr;
 }
 
-std::string log_vtab_impl::get_table_statement(void)
+std::string log_vtab_impl::get_table_statement()
 {
     std::vector<log_vtab_impl::vtab_column> cols;
     std::vector<log_vtab_impl::vtab_column>::const_iterator iter;
     std::ostringstream oss;
     size_t max_name_len = 15;
 
-    oss << "CREATE TABLE " << this->get_name().to_string() << " (\n"
-        << "  log_line        INTEGER  PRIMARY KEY,\n"
-        << "  log_part        TEXT     COLLATE naturalnocase,\n"
-        << "  log_time        DATETIME,\n"
-        << "  log_actual_time DATETIME HIDDEN,\n"
-        << "  log_idle_msecs  INTEGER,\n"
-        << "  log_level       TEXT     COLLATE loglevel,\n"
-        << "  log_mark        BOOLEAN,\n"
-        << "  -- BEGIN Format-specific fields:\n";
+    oss << "CREATE TABLE " << this->get_name().to_string() << LOG_COLUMNS;
     this->get_columns(cols);
     this->vi_column_count = cols.size();
     for (iter = cols.begin(); iter != cols.end(); iter++) {
@@ -85,6 +104,8 @@ std::string log_vtab_impl::get_table_statement(void)
         auto_mem<char, sqlite3_free> colname;
         string comment;
 
+        require(!iter->vc_name.empty());
+
         if (!iter->vc_comment.empty()) {
             comment.append(" -- ")
                    .append(iter->vc_comment);
@@ -96,57 +117,58 @@ std::string log_vtab_impl::get_table_statement(void)
                                   colname.in(),
                                   type_to_string(iter->vc_type),
                                   iter->vc_hidden ? "hidden" : "",
-                                  (iter->vc_collator == NULL ||
-                                   iter->vc_collator[0] == '\0') ?
-                                  "BINARY" : iter->vc_collator,
+                                  iter->vc_collator.empty() ?
+                                  "BINARY" : iter->vc_collator.c_str(),
                                   comment.c_str());
         oss << coldecl;
     }
-    oss << "  -- END Format-specific fields\n"
-        << "  log_path        TEXT HIDDEN COLLATE naturalnocase,\n"
-        << "  log_text        TEXT HIDDEN,\n"
-        << "  log_body        TEXT HIDDEN\n"
-        << ");\n";
+    oss << LOG_FOOTER_COLUMNS;
 
     log_debug("log_vtab_impl.get_table_statement() -> %s", oss.str().c_str());
 
     return oss.str();
 }
 
-int log_vtab_impl::logline_value_to_sqlite_type(logline_value::kind_t kind)
+pair<int, unsigned int> log_vtab_impl::logline_value_to_sqlite_type(value_kind_t kind)
 {
     int type = 0;
+    unsigned int subtype = 0;
 
     switch (kind) {
-        case logline_value::VALUE_NULL:
-        case logline_value::VALUE_TEXT:
-        case logline_value::VALUE_JSON:
-        case logline_value::VALUE_STRUCT:
-        case logline_value::VALUE_QUOTED:
-        case logline_value::VALUE_TIMESTAMP:
+        case value_kind_t::VALUE_JSON:
+            type = SQLITE3_TEXT;
+            subtype = 74;
+            break;
+        case value_kind_t::VALUE_NULL:
+        case value_kind_t::VALUE_TEXT:
+        case value_kind_t::VALUE_STRUCT:
+        case value_kind_t::VALUE_QUOTED:
+        case value_kind_t::VALUE_W3C_QUOTED:
+        case value_kind_t::VALUE_TIMESTAMP:
+        case value_kind_t::VALUE_XML:
             type = SQLITE3_TEXT;
             break;
-        case logline_value::VALUE_FLOAT:
+        case value_kind_t::VALUE_FLOAT:
             type = SQLITE_FLOAT;
             break;
-        case logline_value::VALUE_BOOLEAN:
-        case logline_value::VALUE_INTEGER:
+        case value_kind_t::VALUE_BOOLEAN:
+        case value_kind_t::VALUE_INTEGER:
             type = SQLITE_INTEGER;
             break;
-        case logline_value::VALUE_UNKNOWN:
-        case logline_value::VALUE__MAX:
+        case value_kind_t::VALUE_UNKNOWN:
+        case value_kind_t::VALUE__MAX:
             ensure(0);
             break;
     }
-    return type;
+    return make_pair(type, subtype);
 }
 
 struct vtab {
     sqlite3_vtab        base;
     sqlite3 *           db;
-    textview_curses *tc;
-    logfile_sub_source *lss;
-    log_vtab_impl *     vi;
+    textview_curses *tc{nullptr};
+    logfile_sub_source *lss{nullptr};
+    std::shared_ptr<log_vtab_impl> vi;
 };
 
 struct vtab_cursor {
@@ -164,33 +186,27 @@ static int vt_create(sqlite3 *db,
                      sqlite3_vtab **pp_vt,
                      char **pzErr)
 {
-    log_vtab_manager *vm = (log_vtab_manager *)pAux;
+    auto *vm = (log_vtab_manager *)pAux;
     int   rc             = SQLITE_OK;
-    vtab *p_vt;
-
     /* Allocate the sqlite3_vtab/vtab structure itself */
-    p_vt = (vtab *)sqlite3_malloc(sizeof(*p_vt));
+    auto p_vt = std::make_unique<vtab>();
 
-    if (p_vt == NULL) {
-        return SQLITE_NOMEM;
-    }
-
-    memset(&p_vt->base, 0, sizeof(sqlite3_vtab));
     p_vt->db = db;
 
     /* Declare the vtable's structure */
     p_vt->vi = vm->lookup_impl(intern_string::lookup(argv[3]));
-    if (p_vt->vi == NULL) {
+    if (p_vt->vi == nullptr) {
         return SQLITE_ERROR;
     }
     p_vt->tc = vm->get_view();
     p_vt->lss = vm->get_source();
-    rc        = sqlite3_declare_vtab(db, p_vt->vi->get_table_statement().c_str());
+    rc = sqlite3_declare_vtab(db, p_vt->vi->get_table_statement().c_str());
 
     /* Success. Set *pp_vt and return */
-    *pp_vt = &p_vt->base;
+    auto loose_p_vt = p_vt.release();
+    *pp_vt = &loose_p_vt->base;
 
-    log_debug("creating log format table: %s = %p", argv[3], p_vt);
+    log_debug("creating log format table: %s = %p", argv[3], p_vt.get());
 
     return rc;
 }
@@ -199,8 +215,7 @@ static int vt_destructor(sqlite3_vtab *p_svt)
 {
     vtab *p_vt = (vtab *)p_svt;
 
-    /* Free the SQLite structure */
-    sqlite3_free(p_vt);
+    delete p_vt;
 
     return SQLITE_OK;
 }
@@ -235,7 +250,7 @@ static int vt_open(sqlite3_vtab *p_svt, sqlite3_vtab_cursor **pp_cursor)
     *pp_cursor = (sqlite3_vtab_cursor *)p_cur;
 
     p_cur->base.pVtab = p_svt;
-    p_cur->log_cursor.lc_curr_line = vis_line_t(-1);
+    p_cur->log_cursor.lc_curr_line = -1_vl;
     p_cur->log_cursor.lc_end_line = vis_line_t(p_vt->lss->text_line_count());
     p_cur->log_cursor.lc_sub_index = 0;
     vt_next((sqlite3_vtab_cursor *)p_cur);
@@ -270,8 +285,8 @@ static int vt_next(sqlite3_vtab_cursor *cur)
     do {
         log_cursor_latest = vc->log_cursor;
         if (((log_cursor_latest.lc_curr_line % 1024) == 0) &&
-            (log_vtab_progress_callback != NULL &&
-             log_vtab_progress_callback(log_cursor_latest))) {
+            (log_vtab_data.lvd_progress != NULL &&
+             log_vtab_data.lvd_progress(log_cursor_latest))) {
             break;
         }
         done = vt->vi->next(vc->log_cursor, *vt->lss);
@@ -286,8 +301,10 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     vtab *       vt = (vtab *)cur->pVtab;
 
     content_line_t    cl(vt->lss->at(vc->log_cursor.lc_curr_line));
-    logfile *         lf = vt->lss->find(cl);
-    logfile::iterator ll = lf->begin() + cl;
+    uint64_t line_number;
+    auto ld = vt->lss->find_data(cl, line_number);
+    shared_ptr<logfile> lf = (*ld)->get_file();
+    auto ll = lf->begin() + line_number;
 
     require(col >= 0);
 
@@ -302,32 +319,32 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     case VT_COL_PARTITION:
     {
         vis_bookmarks &vb = vt->tc->get_bookmarks();
-        bookmark_vector<vis_line_t> &bv = vb[&textview_curses::BM_PARTITION];
-        bookmark_vector<vis_line_t>::iterator iter;
-        vis_line_t curr_line;
+        bookmark_vector<vis_line_t> &bv = vb[&textview_curses::BM_META];
 
-        curr_line = vis_line_t(vc->log_cursor.lc_curr_line);
-        iter = lower_bound(bv.begin(), bv.end(), curr_line);
-
-        if (bv.empty() || (iter != bv.end() && curr_line < *iter)) {
+        if (bv.empty()) {
             sqlite3_result_null(ctx);
         }
         else {
-            if (iter == bv.end() || *iter != curr_line) {
-                --iter;
-            }
-            content_line_t part_line = vt->lss->at(*iter);
-            std::map<content_line_t, bookmark_metadata> &bm_meta = vt->lss->get_user_bookmark_metadata();
-            std::map<content_line_t, bookmark_metadata>::iterator meta_iter;
+            vis_line_t curr_line(vc->log_cursor.lc_curr_line);
+            auto iter = lower_bound(bv.begin(), bv.end(), curr_line + 1_vl);
 
-            meta_iter = bm_meta.find(part_line);
-            if (meta_iter != bm_meta.end()) {
-                sqlite3_result_text(ctx,
-                                    meta_iter->second.bm_name.c_str(),
-                                    meta_iter->second.bm_name.size(),
-                                    SQLITE_TRANSIENT);
-            }
-            else {
+            if (iter != bv.begin()) {
+                --iter;
+                content_line_t part_line = vt->lss->at(*iter);
+                std::map<content_line_t, bookmark_metadata> &bm_meta = vt->lss->get_user_bookmark_metadata();
+                std::map<content_line_t, bookmark_metadata>::iterator meta_iter;
+
+                meta_iter = bm_meta.find(part_line);
+                if (meta_iter != bm_meta.end() &&
+                    !meta_iter->second.bm_name.empty()) {
+                    sqlite3_result_text(ctx,
+                                        meta_iter->second.bm_name.c_str(),
+                                        meta_iter->second.bm_name.size(),
+                                        SQLITE_TRANSIENT);
+                } else {
+                    sqlite3_result_null(ctx);
+                }
+            } else {
                 sqlite3_result_null(ctx);
             }
         }
@@ -349,7 +366,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
             if (ll->is_time_skewed()) {
                 if (vc->line_values.empty()) {
                     lf->read_full_message(ll, vc->log_msg);
-                    vt->vi->extract(lf, vc->log_msg, vc->line_values);
+                    vt->vi->extract(lf, line_number, vc->log_msg, vc->line_values);
                 }
 
                 struct line_range time_range;
@@ -384,7 +401,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
             content_line_t prev_cl(vt->lss->at(vis_line_t(
                                                    vc->log_cursor.lc_curr_line -
                                                    1)));
-            logfile *         prev_lf = vt->lss->find(prev_cl);
+            shared_ptr<logfile>         prev_lf = vt->lss->find(prev_cl);
             logfile::iterator prev_ll = prev_lf->begin() + prev_cl;
             uint64_t          prev_time, curr_line_time;
 
@@ -414,6 +431,87 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
     }
     break;
 
+        case VT_COL_LOG_COMMENT: {
+            const auto &bm = vt->lss->get_user_bookmark_metadata();
+
+            auto bm_iter = bm.find(vt->lss->at(vc->log_cursor.lc_curr_line));
+            if (bm_iter == bm.end() || bm_iter->second.bm_comment.empty()) {
+                sqlite3_result_null(ctx);
+            } else {
+                const bookmark_metadata &meta = bm_iter->second;
+                sqlite3_result_text(ctx,
+                                    meta.bm_comment.c_str(),
+                                    meta.bm_comment.length(),
+                                    SQLITE_TRANSIENT);
+            }
+            break;
+        }
+
+        case VT_COL_LOG_TAGS: {
+            const map<content_line_t, bookmark_metadata> &bm = vt->lss->get_user_bookmark_metadata();
+
+            auto bm_iter = bm.find(vt->lss->at(vc->log_cursor.lc_curr_line));
+            if (bm_iter == bm.end() || bm_iter->second.bm_tags.empty()) {
+                sqlite3_result_null(ctx);
+            } else {
+                const bookmark_metadata &meta = bm_iter->second;
+
+                yajlpp_gen gen;
+
+                yajl_gen_config(gen, yajl_gen_beautify, false);
+
+                {
+                    yajlpp_array arr(gen);
+
+                    for (const auto &str : meta.bm_tags) {
+                        arr.gen(str);
+                    }
+                }
+
+                string_fragment sf = gen.to_string_fragment();
+
+                sqlite3_result_text(ctx,
+                                    sf.data(),
+                                    sf.length(),
+                                    SQLITE_TRANSIENT);
+                sqlite3_result_subtype(ctx, 'J');
+            }
+            break;
+        }
+
+        case VT_COL_FILTERS: {
+            auto &filter_mask = (*ld)->ld_filter_state.lfo_filter_state.tfs_mask;
+
+            if (!filter_mask[line_number]) {
+                sqlite3_result_null(ctx);
+            } else {
+                auto &filters = vt->lss->get_filters();
+                yajlpp_gen gen;
+
+                yajl_gen_config(gen, yajl_gen_beautify, false);
+
+                {
+                    yajlpp_array arr(gen);
+
+                    for (auto &filter : filters) {
+                        if (filter->lf_deleted) {
+                            continue;
+                        }
+
+                        uint32_t mask = (1UL << filter->get_index());
+
+                        if (filter_mask[line_number] & mask) {
+                            arr.gen(filter->get_index());
+                        }
+                    }
+                }
+
+                to_sqlite(ctx, gen.to_string_fragment());
+                sqlite3_result_subtype(ctx, 'J');
+            }
+            break;
+        }
+
     default:
         if (col > (VT_COL_MAX + vt->vi->vi_column_count - 1)) {
             int post_col_number = col -
@@ -422,6 +520,10 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
 
             switch (post_col_number) {
                 case 0: {
+                    sqlite3_result_int64(ctx, ll->get_time_in_millis());
+                    break;
+                }
+                case 1: {
                     const string &fn = lf->get_filename();
 
                     sqlite3_result_text(ctx,
@@ -430,7 +532,7 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                                         SQLITE_STATIC);
                     break;
                 }
-                case 1: {
+                case 2: {
                     shared_buffer_ref line;
 
                     lf->read_full_message(ll, line);
@@ -440,19 +542,16 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                                         SQLITE_TRANSIENT);
                     break;
                 }
-                case 2: {
+                case 3: {
                     if (vc->line_values.empty()) {
-                        logfile::iterator line_iter;
-
-                        line_iter = lf->begin() + cl;
-                        lf->read_full_message(line_iter, vc->log_msg);
-                        vt->vi->extract(lf, vc->log_msg, vc->line_values);
+                        lf->read_full_message(ll, vc->log_msg);
+                        vt->vi->extract(lf, line_number, vc->log_msg, vc->line_values);
                     }
 
                     struct line_range body_range;
 
                     body_range = find_string_attr_range(
-                        vt->vi->vi_attrs, &textview_curses::SA_BODY);
+                        vt->vi->vi_attrs, &SA_BODY);
                     if (!body_range.is_valid()) {
                         sqlite3_result_null(ctx);
                     }
@@ -466,15 +565,28 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                     }
                     break;
                 }
+                case 4: {
+                    auto read_res = lf->read_raw_message(ll);
+
+                    if (read_res.isErr()) {
+                        auto msg = fmt::format("unable to read line -- {}",
+                                               read_res.unwrapErr());
+                        sqlite3_result_error(ctx, msg.c_str(), msg.length());
+                    } else {
+                        auto sbr = read_res.unwrap();
+
+                        sqlite3_result_text(ctx,
+                                            sbr.get_data(), sbr.length(),
+                                            SQLITE_TRANSIENT);
+                    }
+                    break;
+                }
             }
         }
         else {
             if (vc->line_values.empty()) {
-                logfile::iterator line_iter;
-
-                line_iter = lf->begin() + cl;
-                lf->read_full_message(line_iter, vc->log_msg);
-                vt->vi->extract(lf, vc->log_msg, vc->line_values);
+                lf->read_full_message(ll, vc->log_msg);
+                vt->vi->extract(lf, line_number, vc->log_msg, vc->line_values);
             }
 
             size_t sub_col = col - VT_COL_MAX;
@@ -484,64 +596,122 @@ static int vt_column(sqlite3_vtab_cursor *cur, sqlite3_context *ctx, int col)
                               logline_value_cmp(NULL, sub_col));
 
             if (lv_iter != vc->line_values.end()) {
-                switch (lv_iter->lv_kind) {
-                case logline_value::VALUE_NULL:
-                    sqlite3_result_null(ctx);
-                    break;
-                case logline_value::VALUE_JSON:
-                case logline_value::VALUE_STRUCT:
-                case logline_value::VALUE_TEXT:
-                case logline_value::VALUE_TIMESTAMP: {
+                if (!lv_iter->lv_meta.lvm_struct_name.empty()) {
+                    yajlpp_gen gen;
+                    yajl_gen_config(gen, yajl_gen_beautify, false);
+
+                    {
+                        yajlpp_map root(gen);
+
+                        for (auto &lv_struct : vc->line_values) {
+                            if (lv_struct.lv_meta.lvm_column != sub_col) {
+                                continue;
+                            }
+
+                            root.gen(lv_struct.lv_meta.lvm_name);
+                            switch (lv_struct.lv_meta.lvm_kind) {
+                                case value_kind_t::VALUE_NULL:
+                                    root.gen();
+                                    break;
+                                case value_kind_t::VALUE_BOOLEAN:
+                                    root.gen((bool) lv_struct.lv_value.i);
+                                    break;
+                                case value_kind_t::VALUE_INTEGER:
+                                    root.gen(lv_struct.lv_value.i);
+                                    break;
+                                case value_kind_t::VALUE_FLOAT:
+                                    root.gen(lv_struct.lv_value.d);
+                                    break;
+                                default:
+                                    root.gen(lv_struct.to_string());
+                                    break;
+                            }
+                        }
+                    }
+
+                    auto sf = gen.to_string_fragment();
                     sqlite3_result_text(ctx,
-                                        lv_iter->text_value(),
-                                        lv_iter->text_length(),
+                                        sf.data(),
+                                        sf.length(),
                                         SQLITE_TRANSIENT);
-                    break;
-                }
-                case logline_value::VALUE_QUOTED:
-                    if (lv_iter->lv_sbr.length() == 0) {
-                        sqlite3_result_text(ctx, "", 0, SQLITE_STATIC);
-                    }
-                    else {
-                        const char *text_value = lv_iter->lv_sbr.get_data();
-                        size_t text_len = lv_iter->lv_sbr.length();
-
-                        switch (text_value[0]) {
-                        case '\'':
-                        case '"': {
-                            char *val = (char *)sqlite3_malloc(text_len);
-
-                            if (val == NULL) {
-                                sqlite3_result_error_nomem(ctx);
-                            }
-                            else {
-                                size_t unquoted_len = unquote(val, text_value, text_len);
-                                sqlite3_result_text(ctx, val, unquoted_len, sqlite3_free);
-                            }
+                    sqlite3_result_subtype(ctx, 74);
+                } else {
+                    switch (lv_iter->lv_meta.lvm_kind) {
+                        case value_kind_t::VALUE_NULL:
+                            sqlite3_result_null(ctx);
+                            break;
+                        case value_kind_t::VALUE_JSON: {
+                            sqlite3_result_text(ctx,
+                                                lv_iter->text_value(),
+                                                lv_iter->text_length(),
+                                                SQLITE_TRANSIENT);
+                            sqlite3_result_subtype(ctx, 74);
                             break;
                         }
-                        default: {
-                            sqlite3_result_text(ctx, text_value,
-                                lv_iter->lv_sbr.length(), SQLITE_TRANSIENT);
+                        case value_kind_t::VALUE_STRUCT:
+                        case value_kind_t::VALUE_TEXT:
+                        case value_kind_t::VALUE_XML:
+                        case value_kind_t::VALUE_TIMESTAMP: {
+                            sqlite3_result_text(ctx,
+                                                lv_iter->text_value(),
+                                                lv_iter->text_length(),
+                                                SQLITE_TRANSIENT);
                             break;
                         }
-                        }
+                        case value_kind_t::VALUE_W3C_QUOTED:
+                        case value_kind_t::VALUE_QUOTED:
+                            if (lv_iter->lv_sbr.empty()) {
+                                sqlite3_result_text(ctx, "", 0, SQLITE_STATIC);
+                            } else {
+                                const char *text_value = lv_iter->lv_sbr.get_data();
+                                size_t text_len = lv_iter->lv_sbr.length();
+
+                                switch (text_value[0]) {
+                                    case '\'':
+                                    case '"': {
+                                        char *val = (char *) sqlite3_malloc(
+                                            text_len);
+
+                                        if (val == nullptr) {
+                                            sqlite3_result_error_nomem(ctx);
+                                        } else {
+                                            auto unquote_func =
+                                                lv_iter->lv_meta.lvm_kind ==
+                                                value_kind_t::VALUE_W3C_QUOTED ?
+                                                unquote_w3c : unquote;
+
+                                            size_t unquoted_len = unquote_func(
+                                                val, text_value, text_len);
+                                            sqlite3_result_text(ctx, val,
+                                                                unquoted_len,
+                                                                sqlite3_free);
+                                        }
+                                        break;
+                                    }
+                                    default: {
+                                        sqlite3_result_text(ctx, text_value,
+                                                            lv_iter->lv_sbr.length(),
+                                                            SQLITE_TRANSIENT);
+                                        break;
+                                    }
+                                }
+                            }
+                            break;
+
+                        case value_kind_t::VALUE_BOOLEAN:
+                        case value_kind_t::VALUE_INTEGER:
+                            sqlite3_result_int64(ctx, lv_iter->lv_value.i);
+                            break;
+
+                        case value_kind_t::VALUE_FLOAT:
+                            sqlite3_result_double(ctx, lv_iter->lv_value.d);
+                            break;
+
+                        case value_kind_t::VALUE_UNKNOWN:
+                        case value_kind_t::VALUE__MAX:
+                            require(0);
+                            break;
                     }
-                    break;
-
-                case logline_value::VALUE_BOOLEAN:
-                case logline_value::VALUE_INTEGER:
-                    sqlite3_result_int64(ctx, lv_iter->lv_value.i);
-                    break;
-
-                case logline_value::VALUE_FLOAT:
-                    sqlite3_result_double(ctx, lv_iter->lv_value.d);
-                    break;
-
-                case logline_value::VALUE_UNKNOWN:
-                case logline_value::VALUE__MAX:
-                    require(0);
-                    break;
                 }
             }
             else {
@@ -567,7 +737,7 @@ static int vt_rowid(sqlite3_vtab_cursor *cur, sqlite_int64 *p_rowid)
 void log_cursor::update(unsigned char op, vis_line_t vl, bool exact)
 {
     if (vl < 0) {
-        vl = vis_line_t(-1);
+        vl = -1_vl;
     }
     switch (op) {
     case SQLITE_INDEX_CONSTRAINT_EQ:
@@ -601,7 +771,7 @@ static int vt_filter(sqlite3_vtab_cursor *p_vtc,
         sqlite3_index_info::sqlite3_index_constraint *)idxStr;
 
     log_info("(%p) filter called: %d", vt, idxNum);
-    p_cur->log_cursor.lc_curr_line = vis_line_t(-1);
+    p_cur->log_cursor.lc_curr_line = -1_vl;
     p_cur->log_cursor.lc_end_line = vis_line_t(vt->lss->text_line_count());
     vt_next(p_vtc);
 
@@ -616,28 +786,29 @@ static int vt_filter(sqlite3_vtab_cursor *p_vtc,
                 vis_line_t(sqlite3_value_int64(argv[lpc])));
             break;
 
-        case VT_COL_LOG_TIME: {
-            const unsigned char *datestr = sqlite3_value_text(argv[lpc]);
-            date_time_scanner dts;
-            struct timeval tv;
-            struct exttm mytm;
-            vis_line_t vl;
+        case VT_COL_LOG_TIME:
+            if (sqlite3_value_type(argv[lpc]) == SQLITE3_TEXT) {
+                const unsigned char *datestr = sqlite3_value_text(argv[lpc]);
+                date_time_scanner dts;
+                struct timeval tv;
+                struct exttm mytm;
+                vis_line_t vl;
 
-            dts.scan((const char *)datestr, strlen((const char *)datestr), NULL, &mytm, tv);
-            if ((vl = vt->lss->find_from_time(tv)) == -1) {
-                p_cur->log_cursor.lc_curr_line = p_cur->log_cursor.lc_end_line;
-            }
-            else {
-                p_cur->log_cursor.update(index[lpc].op, vl, false);
+                dts.scan((const char *)datestr, strlen((const char *)datestr), NULL, &mytm, tv);
+                if ((vl = vt->lss->find_from_time(tv)) == -1) {
+                    p_cur->log_cursor.lc_curr_line = p_cur->log_cursor.lc_end_line;
+                }
+                else {
+                    p_cur->log_cursor.update(index[lpc].op, vl, false);
+                }
             }
             break;
-        }
 
         }
     }
 
     while (!p_cur->log_cursor.is_eof() && !vt->vi->is_valid(p_cur->log_cursor, *vt->lss)) {
-        p_cur->log_cursor.lc_curr_line += vis_line_t(1);
+        p_cur->log_cursor.lc_curr_line += 1_vl;
     }
 
     return SQLITE_OK;
@@ -706,6 +877,14 @@ static int vt_best_index(sqlite3_vtab *tab, sqlite3_index_info *p_info)
     return SQLITE_OK;
 }
 
+static struct json_path_container tags_handler = {
+    json_path_handler("#")
+        .with_synopsis("<tag>")
+        .with_description("A tag for the log line")
+        .with_pattern(R"(^#[^\s]+$)")
+        .FOR_FIELD(bookmark_metadata, bm_tags)
+};
+
 static int vt_update(sqlite3_vtab *tab,
                      int argc,
                      sqlite3_value **argv,
@@ -722,52 +901,95 @@ static int vt_update(sqlite3_vtab *tab,
 
         std::map<content_line_t, bookmark_metadata> &bm = vt->lss->get_user_bookmark_metadata();
         const unsigned char *part_name = sqlite3_value_text(argv[2 + VT_COL_PARTITION]);
+        const unsigned char *log_comment = sqlite3_value_text(argv[2 + VT_COL_LOG_COMMENT]);
+        const unsigned char *log_tags = sqlite3_value_text(argv[2 + VT_COL_LOG_TAGS]);
+        bookmark_metadata tmp_bm;
+
+        if (log_tags) {
+            vector<string> errors;
+            yajlpp_parse_context ypc(log_vtab_data.lvd_source, &tags_handler);
+            auto_mem<yajl_handle_t> handle(yajl_free);
+
+            handle = yajl_alloc(&ypc.ypc_callbacks, nullptr, &ypc);
+            ypc.ypc_userdata = &errors;
+            ypc.ypc_line_number = log_vtab_data.lvd_line_number;
+            ypc.with_handle(handle)
+                .with_error_reporter([](const yajlpp_parse_context &ypc,
+                                        lnav_log_level_t level,
+                                        const char *msg) {
+                    vector<string> &errors = *((vector<string> *) ypc.ypc_userdata);
+                    errors.emplace_back(msg);
+                })
+                .with_obj(tmp_bm);
+            ypc.parse(log_tags, strlen((const char *) log_tags));
+            ypc.complete_parse();
+            if (!errors.empty()) {
+                auto all_errors = fmt::format("{}", fmt::join(errors, "\n"));
+                tab->zErrMsg = sqlite3_mprintf("%s", all_errors.c_str());
+                return SQLITE_ERROR;
+            }
+        }
 
         bookmark_vector<vis_line_t> &bv = vt->tc->get_bookmarks()[
-                &textview_curses::BM_PARTITION];
-        bookmark_vector<vis_line_t>::iterator part_iter;
-        bool set_name = false;
+                &textview_curses::BM_META];
+        bool has_meta = part_name != nullptr || log_comment != nullptr ||
+            log_tags != nullptr;
 
-        if ((part_iter = find(bv.begin(), bv.end(), vrowid)) != bv.end()) {
-            if (part_name == NULL) {
-                vt->tc->set_user_mark(&textview_curses::BM_PARTITION, vrowid, false);
-                bm.erase(vt->lss->at(vrowid));
-            }
-            else {
-                set_name = true;
-            }
-        }
-        else if (part_name != NULL) {
-            vt->tc->set_user_mark(&textview_curses::BM_PARTITION, vrowid, true);
-            set_name = true;
+        if (binary_search(bv.begin(), bv.end(), vrowid) && !has_meta) {
+            vt->tc->set_user_mark(&textview_curses::BM_META, vrowid, false);
+            bm.erase(vt->lss->at(vrowid));
         }
 
-        if (set_name) {
+        if (has_meta) {
             bookmark_metadata &line_meta = bm[vt->lss->at(vrowid)];
 
-            line_meta.bm_name = string((const char *) part_name);
+            vt->tc->set_user_mark(&textview_curses::BM_META, vrowid, true);
+            if (part_name) {
+                line_meta.bm_name = string((const char *) part_name);
+            } else {
+                line_meta.bm_name.clear();
+            }
+            if (log_comment) {
+                line_meta.bm_comment = string((const char *) log_comment);
+            } else {
+                line_meta.bm_comment.clear();
+            }
+            if (log_tags) {
+                line_meta.bm_tags.clear();
+                for (const auto &tag : tmp_bm.bm_tags) {
+                    line_meta.add_tag(tag);
+                }
+
+                for (const auto &tag : line_meta.bm_tags) {
+                    bookmark_metadata::KNOWN_TAGS.insert(tag);
+                }
+            } else {
+                line_meta.bm_tags.clear();
+            }
         }
 
-        vt->tc->set_user_mark(&textview_curses::BM_USER, vis_line_t(rowid), val);
+        vt->tc->set_user_mark(&textview_curses::BM_USER, vrowid, val);
         rowid += 1;
         while ((size_t)rowid < vt->lss->text_line_count()) {
             vis_line_t vl(rowid);
             content_line_t cl = vt->lss->at(vl);
             logline *ll = vt->lss->find_line(cl);
-            if (!ll->is_continued()) {
+            if (ll->is_message()) {
                 break;
             }
             vt->tc->set_user_mark(&textview_curses::BM_USER, vl, val);
             rowid += 1;
         }
 
-        retval = SQLITE_OK;
+        if (retval != SQLITE_ERROR) {
+            retval = SQLITE_OK;
+        }
     }
 
     return retval;
 }
 
-static sqlite3_module vtab_module = {
+static sqlite3_module generic_vtab_module = {
     0,              /* iVersion */
     vt_create,      /* xCreate       - create a vtable */
     vt_connect,     /* xConnect      - associate a vtable with a connection */
@@ -793,8 +1015,8 @@ static int progress_callback(void *ptr)
 {
     int retval = 0;
 
-    if (log_vtab_progress_callback != NULL) {
-        retval = log_vtab_progress_callback(log_cursor_latest);
+    if (log_vtab_data.lvd_progress != nullptr) {
+        retval = log_vtab_data.lvd_progress(log_cursor_latest);
     }
 
     return retval;
@@ -805,17 +1027,26 @@ log_vtab_manager::log_vtab_manager(sqlite3 *memdb,
                                    logfile_sub_source &lss)
     : vm_db(memdb), vm_textview(tc), vm_source(lss)
 {
-    sqlite3_create_module(this->vm_db, "log_vtab_impl", &vtab_module, this);
-    sqlite3_progress_handler(memdb, 32, progress_callback, NULL);
+    sqlite3_create_module(this->vm_db, "log_vtab_impl", &generic_vtab_module, this);
+    sqlite3_progress_handler(memdb, 32, progress_callback, nullptr);
 }
 
-string log_vtab_manager::register_vtab(log_vtab_impl *vi)
+log_vtab_manager::~log_vtab_manager()
+{
+    while (!this->vm_impls.empty()) {
+        auto first_name = this->vm_impls.begin()->first;
+
+        this->unregister_vtab(first_name);
+    }
+}
+
+string log_vtab_manager::register_vtab(std::shared_ptr<log_vtab_impl> vi)
 {
     string retval;
 
     if (this->vm_impls.find(vi->get_name()) == this->vm_impls.end()) {
-        auto_mem<char> errmsg(sqlite3_free);
-        auto_mem<char> sql(sqlite3_free);
+        auto_mem<char, sqlite3_free> errmsg;
+        auto_mem<char, sqlite3_free> sql;
         int            rc;
 
         this->vm_impls[vi->get_name()] = vi;
@@ -826,8 +1057,8 @@ string log_vtab_manager::register_vtab(log_vtab_impl *vi)
                               vi->get_name().get());
         rc = sqlite3_exec(this->vm_db,
                           sql,
-                          NULL,
-                          NULL,
+                          nullptr,
+                          nullptr,
                           errmsg.out());
         if (rc != SQLITE_OK) {
             retval = errmsg;
@@ -848,7 +1079,8 @@ string log_vtab_manager::unregister_vtab(intern_string_t name)
         retval = "unknown log line table -- " + name.to_string();
     }
     else {
-        auto_mem<char> sql(sqlite3_free);
+        auto_mem<char, sqlite3_free> sql;
+        __attribute((unused))
         int   rc;
 
         sql = sqlite3_mprintf("DROP TABLE %s ", name.get());
