@@ -1,5 +1,30 @@
 #! /bin/bash
 
+run_test ./drive_sql "select humanize_file_size()"
+
+check_error_output "" <<EOF
+error: sqlite3_exec failed -- wrong number of arguments to function humanize_file_size()
+EOF
+
+run_test ./drive_sql "select humanize_file_size('abc')"
+
+check_error_output "" <<EOF
+error: sqlite3_exec failed -- Expecting an integer for argument number 0
+EOF
+
+run_test ./drive_sql "select humanize_file_size(1, 2)"
+
+check_error_output "" <<EOF
+error: sqlite3_exec failed -- wrong number of arguments to function humanize_file_size()
+EOF
+
+run_test ./drive_sql "select humanize_file_size(10 * 1000 * 1000)"
+
+check_output "" <<EOF
+Row 0:
+  Column humanize_file_size(10 * 1000 * 1000): 9.5MB
+EOF
+
 run_test ./drive_sql "select startswith('.foo', '.')"
 
 check_output "" <<EOF
@@ -77,6 +102,41 @@ Row 0:
   Column regexp_replace('test 1 2 3', '\d+', 'N'): test N N N
 EOF
 
+run_test ./drive_sql "select regexp_replace('test 1 2 3', '\\s+', '{\\0}') as repl"
+
+check_output "" <<EOF
+Row 0:
+  Column       repl: test{ }1{ }2{ }3
+EOF
+
+run_test ./drive_sql "select regexp_replace('test 1 2 3', '\\w*', '{\\0}') as repl"
+
+check_output "" <<EOF
+Row 0:
+  Column       repl: {test}{} {1}{} {2}{} {3}{}
+EOF
+
+run_test ./drive_sql "select regexp_replace('123 abc', '(\w*)', '<\3>') as repl"
+
+check_output "" <<EOF
+Row 0:
+  Column       repl: <\3><\3> <\3><\3>
+EOF
+
+run_test ./drive_sql "select regexp_replace('123 abc', '(\w*)', '<\\\\>') as repl"
+
+check_output "" <<EOF
+Row 0:
+  Column       repl: <\><\> <\><\>
+EOF
+
+run_test ./drive_sql "select regexp_replace('abc: def', '(\w*):\s*(.*)', '\1=\2') as repl"
+
+check_output "" <<EOF
+Row 0:
+  Column       repl: abc=def
+EOF
+
 
 run_test ./drive_sql "select regexp_match('abc', 'abc')"
 
@@ -174,6 +234,13 @@ Row 0:
 EOF
 
 
+run_test ./drive_sql "SELECT substr('#foo', range_start) AS value FROM regexp_capture('#foo', '(\w+)') WHERE capture_index = 1"
+
+check_output "" <<EOF
+ Row 0:
+  Column      value: foo
+EOF
+
 run_test ./drive_sql "SELECT * FROM regexp_capture('foo bar', '\w+ (\w+)')"
 
 check_output "" <<EOF
@@ -182,16 +249,16 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 0
-  Column range_stop: 7
+  Column range_start: 1
+  Column range_stop: 8
   Column    content: foo bar
 Row 1:
   Column match_index: 0
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 2
-  Column range_start: 4
-  Column range_stop: 7
+  Column range_start: 5
+  Column range_stop: 8
   Column    content: bar
 EOF
 
@@ -203,8 +270,8 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 1
-  Column range_start: 0
-  Column range_stop: 7
+  Column range_start: 1
+  Column range_stop: 8
   Column    content: foo bar
 EOF
 
@@ -216,16 +283,16 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 0
-  Column range_stop: 7
+  Column range_start: 1
+  Column range_stop: 8
   Column    content: foo bar
 Row 1:
   Column match_index: 0
   Column capture_index: 1
   Column capture_name: word
   Column capture_count: 2
-  Column range_start: 4
-  Column range_stop: 7
+  Column range_start: 5
+  Column range_stop: 8
   Column    content: bar
 EOF
 
@@ -237,24 +304,24 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 3
-  Column range_start: 0
-  Column range_stop: 7
+  Column range_start: 1
+  Column range_stop: 8
   Column    content: foo bar
 Row 1:
   Column match_index: 0
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 3
-  Column range_start: -1
-  Column range_stop: -1
+  Column range_start: 0
+  Column range_stop: 0
   Column    content: (null)
 Row 2:
   Column match_index: 0
   Column capture_index: 2
   Column capture_name: word
   Column capture_count: 3
-  Column range_start: 4
-  Column range_stop: 7
+  Column range_start: 5
+  Column range_stop: 8
   Column    content: bar
 EOF
 
@@ -282,64 +349,64 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 0
-  Column range_stop: 1
+  Column range_start: 1
+  Column range_stop: 2
   Column    content: 1
 Row 1:
   Column match_index: 0
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 2
-  Column range_start: 0
-  Column range_stop: 1
+  Column range_start: 1
+  Column range_stop: 2
   Column    content: 1
 Row 2:
   Column match_index: 1
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 2
-  Column range_stop: 3
+  Column range_start: 3
+  Column range_stop: 4
   Column    content: 2
 Row 3:
   Column match_index: 1
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 2
-  Column range_start: 2
-  Column range_stop: 3
+  Column range_start: 3
+  Column range_stop: 4
   Column    content: 2
 Row 4:
   Column match_index: 2
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 4
-  Column range_stop: 5
+  Column range_start: 5
+  Column range_stop: 6
   Column    content: 3
 Row 5:
   Column match_index: 2
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 2
-  Column range_start: 4
-  Column range_stop: 5
+  Column range_start: 5
+  Column range_stop: 6
   Column    content: 3
 Row 6:
   Column match_index: 3
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 2
-  Column range_start: 6
-  Column range_stop: 8
+  Column range_start: 7
+  Column range_stop: 9
   Column    content: 45
 Row 7:
   Column match_index: 3
   Column capture_index: 1
   Column capture_name:
   Column capture_count: 2
-  Column range_start: 6
-  Column range_stop: 8
+  Column range_start: 7
+  Column range_stop: 9
   Column    content: 45
 EOF
 
@@ -351,7 +418,7 @@ Row 0:
   Column capture_index: 0
   Column capture_name: (null)
   Column capture_count: 1
-  Column range_start: 0
-  Column range_stop: 3
+  Column range_start: 1
+  Column range_stop: 4
   Column    content: foo
 EOF
